grammar alternating_machine;

options{
	output=template;
	rewrite=true;
	language=Java;
	backtrack=true; 
	memoize=true;
}
@header{
package at.ac.tuwien.dbai.staff.dvorak.alternation.compiler;

import java.util.SortedMap;
import java.util.TreeMap;
import java.util.Set;
import java.util.HashSet;
}
@lexer::header{
package at.ac.tuwien.dbai.staff.dvorak.alternation.compiler;
}
@members{
	List<String> inputVariables = new ArrayList<String>();
	SortedMap<String,String> inputVariablesMap = new TreeMap<String,String>();
	SortedMap<String,String> workVariables = new TreeMap<String,String>();
	List<String> workInitialization = new ArrayList<String>();
	Map<String,Byte> states = new HashMap<String,Byte>();
  	byte countstates=0;
	private byte startStateId=-1;
	Map<String,String> primitiveTypes = new HashMap<String,String>(8);
	private String programName="noName";
	private String ATMType= "cyclic"; // cyclic acceptCyclic acyclic simple 
	
	private void init() {
		primitiveTypes.put("boolean","Boolean");
    		primitiveTypes.put("char","Character");
    		primitiveTypes.put("byte","Byte");
    		primitiveTypes.put("short","Short");
    		primitiveTypes.put("int","Integer");
    		primitiveTypes.put("long","Long");
    		primitiveTypes.put("float","Float");
    		primitiveTypes.put("double","Double");
	}
		
	private String inputTape(){
		
		List<String> TypeInputVariables=new ArrayList<String>();
		for(String entry: inputVariables) {
			TypeInputVariables.add(inputVariablesMap.get(entry)+" "+entry);	
		}
		StringTemplate inputtape=templateLib.getInstanceOf("inputtape");
		inputtape.setAttribute("vars",inputVariables);
		inputtape.setAttribute("typeVars",TypeInputVariables);
		return inputtape.toString();
	}
	
	private String workTape(){
	
		List<String> TypeWorkVariables=new ArrayList<String>();
		List<String> vars=new ArrayList<String>();
		List<String> primitiveVars=new ArrayList<String>();
		List<String> arrays =new ArrayList<String>();
		
		StringTemplate template;
		String value;
		String key;
		for(Map.Entry<String,String> entry: workVariables.entrySet()) {
			TypeWorkVariables.add(entry.getValue()+" "+entry.getKey());
		
			if(primitiveTypes.containsKey(entry.getValue())){
				primitiveVars.add(entry.getKey());					
											
				continue;
			}  
			
			if(entry.getValue().contains("[]")){
				arrays.add(entry.getKey());
				
				continue;
			}
			vars.add(entry.getKey());			
		}

		StringTemplate worktape=templateLib.getInstanceOf("worktape");
		worktape.setAttribute("vars",vars);
		worktape.setAttribute("primitiveVars",primitiveVars);
		worktape.setAttribute("arrays",arrays);
		worktape.setAttribute("typeVars",TypeWorkVariables);
		worktape.setAttribute("inits",workInitialization);
		return worktape.toString();
	}
	
	private String compute() {
		List<String> TypeInputVariables=new ArrayList<String>();
		for(String entry: inputVariables) {
			TypeInputVariables.add(inputVariablesMap.get(entry)+" "+entry);	
		}
		StringTemplate compute=templateLib.getInstanceOf("compute");
		compute.setAttribute("inputVars",inputVariables);
		compute.setAttribute("inputTypeVars",TypeInputVariables);
		
		List<String> putStates=new ArrayList<String>();
		for(Map.Entry<String,Byte> entry: states.entrySet()){
			putStates.add("atmStates.put((byte)"+entry.getValue()+",new "+entry.getKey()+"());");
		}
		compute.setAttribute("putStates",putStates);
		compute.setAttribute("startStateId",startStateId);
		compute.setAttribute("atmtype",ATMType);		
				
		return compute.toString();
	}
	
	private String getComputationTree() {
		StringTemplate stringTemplate=templateLib.getInstanceOf("getcomputationtree");
		List<String> idToName=new ArrayList<String>();
		for(Map.Entry<String,Byte> entry: states.entrySet()){
			idToName.add("cT.setStateIdToName((byte)"+entry.getValue()+", \""+entry.getKey()+"\");");
			idToName.add("cT.setNameToStateId(\""+entry.getKey()+"\",(byte) "+entry.getValue()+");");
		}
		stringTemplate.setAttribute("idToName",idToName);
		stringTemplate.setAttribute("name",programName);		

		return stringTemplate.toString();
	
	}
	
	
	private void newState(String name) {	
		if(!states.containsKey(name)) {
			states.put(name,countstates);
			countstates++;
		}
		if(startStateId==-1) {
			startStateId=states.get(name);
		}
		
	}
	
	private byte getStateNumber(String name) {
		if(!states.containsKey(name)) {
			states.put(name,countstates);
			countstates++;
		}
		return states.get(name);		
	}
	
	private String getPrefix(String Identifier) {
		if (workVariables.containsKey(Identifier)){
			return "atmWorktape.";
		}
		
		if (inputVariables.contains(Identifier)){
			return "atmInputtape.";
		}
		
		return "";
	}
}


atm_file  :{init();}packageDecl
	imp+=importDeclaration*
	atm
               ;
	  
atm	 :	head
		'{'
			worktape
			states+=state+
		'}'
		
	-> atmClass(name={$head.st}, inputtape={inputTape()},worktape={workTape()},states={$states},compute={compute()},getcomputationtree={getComputationTree()}) 
	;

packageDecl
    :	'package' qualifiedName ';'
    ;
    
importDeclaration
    :   'import' 'static'? qualifiedName ('.' '*')? ';' 
    ;
    
head	: (atmtype=Identifier{ATMType=$atmtype.text;})? 'atm' name=Identifier { programName=$name.text;} '('input')' 
	-> template(id={$name.text}) "<id>"
	;

input	: arg(',' arg)*
	;
	
arg	: type Identifier {inputVariables.add($Identifier.text); inputVariablesMap.put($Identifier.text,$type.text);}
	;	
	
worktape: declaration*
	;

declaration
	: args ';' 
	| primitiveType Identifier  '=' expression';'{workVariables.put($Identifier.text,$primitiveType.text); 
	workInitialization.add($Identifier.text + " = " + $expression.text);} 
	| classOrInterfaceType Identifier  '=' expression';'{workVariables.put($Identifier.text,$classOrInterfaceType.text); 
	workInitialization.add($Identifier.text + " = " + $expression.text);} 
	| oneDimensionalArrayType Identifier '=' ( 'new' arrayCreator ';'{workVariables.put($Identifier.text,$oneDimensionalArrayType.text); 
	workInitialization.add($Identifier.text + " = new " + $arrayCreator.text);} |expression';'{workVariables.put($Identifier.text,$oneDimensionalArrayType.text); 
	workInitialization.add($Identifier.text + " = " + $expression.text);} )
	;

oneDimensionalArrayType
	: (primitiveType| classOrInterfaceType)'['']'	
	;
	
args	: classOrInterfaceType id1=Identifier {workVariables.put($id1.text,$classOrInterfaceType.text);} (',' id2=Identifier {workVariables.put($id2.text,$classOrInterfaceType.text);})*
	;
	
state	: 'state' Identifier {newState($Identifier.text);}  '{' atmJava '}'
	-> state(id={$Identifier.text}, atmJava={$atmJava.text})
	;
atmJava
	:stmts+=statement+ //->atmJava(stmts={$stmts})
	;	
/*
 [The "BSD licence"]
 Copyright (c) 2007-2008 Terence Parr
 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.
 3. The name of the author may not be used to endorse or promote products
    derived from this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
	
typeList
    :   type (',' type)*
    ;
                             
variableDeclarators
    :   variableDeclarator (',' variableDeclarator)*
    ;

variableDeclarator
    :   variableDeclaratorId ('=' variableInitializer)?
    ;
    
/*
constantDeclarator
    :   Identifier constantDeclaratorRest
    ;
    
constantDeclaratorsRest
    :   constantDeclaratorRest (',' constantDeclarator)*
    ;

constantDeclaratorRest
    :   ('[' ']')* '=' variableInitializer
    ;
*/  
  
variableDeclaratorId
    :   Identifier ('[' ']')*
    ;

variableInitializer
    :   arrayInitializer
    |   expression
    ;
        
arrayInitializer
    :   '{' (variableInitializer (',' variableInitializer)* (',')? )? '}'
    ;	


type
	:	classOrInterfaceType ('[' ']')*
	|	primitiveType ('[' ']')*
	;


classOrInterfaceType
	:	Identifier typeArguments? ('.' Identifier typeArguments? )*
	;
	
primitiveType
    :   'boolean'
    |   'char'
    |   'byte'
    |   'short'
    |   'int'
    |   'long'
    |   'float'
    |   'double'
    ;

        
typeArguments
    :   '<' typeArgument (',' typeArgument)* '>'
    ;
    
typeArgument
    :   type
    |   '?' (('extends' | 'super') type)?
    ;


qualifiedName
    :   Identifier('.' Identifier)*
    ;

literal 
    :   integerLiteral
    |   FloatingPointLiteral
    |   CharacterLiteral
    |   StringLiteral
    |   booleanLiteral
    |   'null'
    ;

integerLiteral
    :   HexLiteral
    |   OctalLiteral
    |   DecimalLiteral
    ;

booleanLiteral
    :   'true'  
    |   'false'
    ;

// STATEMENTS / BLOCKS

block
    :   '{' stmts+=blockStatement* '}' 
    ;
    
blockStatement
    :	statement 
    ;
    
localVariableDeclaration
    :   type variableDeclarators //done
    ;
    

statement
    : block 
    |   'if' parExpression statement (options {k=1;}:'else' statement)?
    |   'for' '(' forControl ')' statement 
    |   'while' parExpression statement 
    |   'do' statement 'while' parExpression ';' 
    |   'try' block
        ( catches 'finally' block
        | catches 
        |  'finally' block 
        )
    |   'switch' parExpression '{' switchBlockStatementGroups '}' 
    |   'throw' expression ';'
    |   'break' (Identifier)? ';' 
    |   'continue' (Identifier)?  ';'
    |   ';' 
    |   statementExpression ';' 
    |   Identifier ':' statement 
    |   ACCEPT ';' -> accept()
    |   REJECT ';' -> reject()
    |   FORALL '{' nondeterminism '}' -> forall(nondeterminism={$nondeterminism.text})
    |   EXISTS '{' nondeterminism '}' -> exists(nondeterminism={$nondeterminism.text})
    ;
    
catches
    :   (catchClause)+
    ;
    
catchClause
    :   'catch' '(' formalParameter ')' block
    ;

formalParameter
    :  type variableDeclaratorId//done
    ;
        
switchBlockStatementGroups
    :   (switchBlockStatementGroup)* 
    ;
    
/* The change here (switchLabel -> switchLabel+) technically makes this grammar
   ambiguous; but with appropriately greedy parsing it yields the most
   appropriate AST, one in which each group, except possibly the last one, has
   labels and statements. */
switchBlockStatementGroup
    :   sL+=switchLabel+ bS+=blockStatement*
    ;
    
switchLabel
    :   'case' constantExpression ':' 
    |   'default' ':' 
    ;
    
forControl
options {k=3;} // be efficient for common case: for (ID ID : ID) ...
    :   enhancedForControl
    |   forInit? ';' expression? ';' forUpdate? 
    ;

forInit
    :   localVariableDeclaration 
    |   expressionList 
    ;
    
enhancedForControl
    :   type Identifier ':' expression
    ;

forUpdate
    :   expressionList 
    ;

// EXPRESSIONS

parExpression
    :   '(' expression ')' 
    ;
    
expressionList
    :   expression (',' expression)* 
    ;

statementExpression
    :   expression
    ;
    
constantExpression
    :   expression
    ;
    
expression
    :   conditionalExpression (assignmentOperator expression)? 
    ;
    
assignmentOperator
    :   '='
    |   '+='
    |   '-='
    |   '*='
    |   '/='
    |   '&='
    |   '|='
    |   '^='
    |   '%='
    |   ('<' '<' '=')=> t1='<' t2='<' t3='=' 
        { $t1.getLine() == $t2.getLine() &&
          $t1.getCharPositionInLine() + 1 == $t2.getCharPositionInLine() && 
          $t2.getLine() == $t3.getLine() && 
          $t2.getCharPositionInLine() + 1 == $t3.getCharPositionInLine() }?
    |   ('>' '>' '>' '=')=> t1='>' t2='>' t3='>' t4='='
        { $t1.getLine() == $t2.getLine() && 
          $t1.getCharPositionInLine() + 1 == $t2.getCharPositionInLine() &&
          $t2.getLine() == $t3.getLine() && 
          $t2.getCharPositionInLine() + 1 == $t3.getCharPositionInLine() &&
          $t3.getLine() == $t4.getLine() && 
          $t3.getCharPositionInLine() + 1 == $t4.getCharPositionInLine() }?
    |   ('>' '>' '=')=> t1='>' t2='>' t3='='
        { $t1.getLine() == $t2.getLine() && 
          $t1.getCharPositionInLine() + 1 == $t2.getCharPositionInLine() && 
          $t2.getLine() == $t3.getLine() && 
          $t2.getCharPositionInLine() + 1 == $t3.getCharPositionInLine() }?
    ;

conditionalExpression
    :   conditionalOrExpression ( '?' expression ':' expression )?
    ;

conditionalOrExpression
    :   conditionalAndExpression ( '||' conditionalAndExpression )*
    ;

conditionalAndExpression
    :   inclusiveOrExpression ( '&&' inclusiveOrExpression )*
    ;

inclusiveOrExpression
    :   exclusiveOrExpression ( '|' exclusiveOrExpression )*
    ;

exclusiveOrExpression
    :   andExpression ( '^' andExpression )*
    ;

andExpression
    :   equalityExpression ( '&' equalityExpression )*
    ;

equalityExpression
    :   instanceOfExpression ( ('==' | '!=') instanceOfExpression )*
    ;

instanceOfExpression
    :   relationalExpression ('instanceof' type)?
    ;

relationalExpression
    :   shiftExpression ( relationalOp shiftExpression )*
    ;
    
relationalOp
    :   ('<' '=')=> t1='<' t2='=' 
        { $t1.getLine() == $t2.getLine() && 
          $t1.getCharPositionInLine() + 1 == $t2.getCharPositionInLine() }?
    |   ('>' '=')=> t1='>' t2='=' 
        { $t1.getLine() == $t2.getLine() && 
          $t1.getCharPositionInLine() + 1 == $t2.getCharPositionInLine() }?
    |   '<' 
    |   '>' 
    ;

shiftExpression
    :   additiveExpression ( shiftOp additiveExpression )*
    ;

shiftOp
    :   ('<' '<')=> t1='<' t2='<' 
        { $t1.getLine() == $t2.getLine() && 
          $t1.getCharPositionInLine() + 1 == $t2.getCharPositionInLine() }?
    |   ('>' '>' '>')=> t1='>' t2='>' t3='>' 
        { $t1.getLine() == $t2.getLine() && 
          $t1.getCharPositionInLine() + 1 == $t2.getCharPositionInLine() &&
          $t2.getLine() == $t3.getLine() && 
          $t2.getCharPositionInLine() + 1 == $t3.getCharPositionInLine() }?
    |   ('>' '>')=> t1='>' t2='>'
        { $t1.getLine() == $t2.getLine() && 
          $t1.getCharPositionInLine() + 1 == $t2.getCharPositionInLine() }?
    ;


additiveExpression
    :   multiplicativeExpression ( ('+' | '-') multiplicativeExpression )*
    ;

multiplicativeExpression
    :   unaryExpression ( ( '*' | '/' | '%' ) unaryExpression )*
    ;
    
unaryExpression
    :   '+' unaryExpression
    |   '-' unaryExpression
    |   '++' unaryExpression
    |   '--' unaryExpression
    |   unaryExpressionNotPlusMinus
    ;

unaryExpressionNotPlusMinus
    :   '~' unaryExpression
    |   '!' unaryExpression
    |   castExpression
    |   primary selector* ('++'|'--')?
    ;

castExpression
    :  '(' primitiveType ')' unaryExpression
    |  '(' (type | expression) ')' unaryExpressionNotPlusMinus
    ;

primary
    :   parExpression
    |   literal
    |   id=Identifier ('.' Identifier)* identifierSuffix? -> template(text={$text},prefix={getPrefix($id.text)}) "<prefix><text>"
    |   primitiveType ('[' ']')* '.' 'class'
    |   'void' '.' 'class'
    //|   'new' creator
    ;

identifierSuffix
    :   ('[' ']')+ '.' 'class'
    |   ('[' expression ']')+ // can also be matched by selector, but do here
    |   arguments
    |   '.' 'class'
    |   '.' explicitGenericInvocation
    ;

arrayCreator
    :  createdName '['
        (  ']' arrayInitializer
        |   expression ']' 
        )
    ;

createdName
    :   classOrInterfaceType
    |   primitiveType
    ;
    
explicitGenericInvocation
    :   nonWildcardTypeArguments Identifier arguments
    ;
    
nonWildcardTypeArguments
    :   '<' typeList '>'
    ;
    
selector
    :   '.' Identifier arguments?
    |   '[' expression ']'
    ;
    
arguments
    :   '(' expressionList? ')'
    ;   

// forall - exists
nondeterminism
    :    (stateCall
    |   stateLoop)+
    ;

stateCall
    : Identifier stateBlock -> stateCall(id={getStateNumber($Identifier.text)},block={$stateBlock.text})
    | 'if' parExpression '{' stateCall+ '}' (options {k=1;}:'else' '{'stateCall+'}')?
    ;

stateLoop
    : 'for' '(' forControl ')' '{' stateCall '}' 
    ;

stateBlock
    :   '{' stateStatement* '}' 
    ;

stateStatement
    : stateBlock  
    |   'if' parExpression stateStatement (options {k=1;}:'else' stateStatement)?
    |   'for' '(' forControl ')' stateStatement  
    |   'while' parExpression stateStatement
    |   'do' stateStatement 'while' parExpression ';'
    |   'try' b1=stateBlock
        ( catches 'finally' stateBlock 
        | catches 
        |   'finally' stateBlock 
        )
    |   'switch' parExpression '{' switchBlockStatementGroups '}'
    |   'throw' expression ';' 
    |   'break' (Identifier)? ';' 	
    |   'continue' (Identifier)?  ';'
    |   ';' 
    |   statementExpression ';'
    |   Identifier ':' statement 
    ;   

// LEXER
ACCEPT	:	'accept';

REJECT	:	'reject';

FORALL	:	'forall';

EXISTS	:	'exists';


HexLiteral : '0' ('x'|'X') HexDigit+ IntegerTypeSuffix? ;

DecimalLiteral : ('0' | '1'..'9' '0'..'9'*) IntegerTypeSuffix? ;

OctalLiteral : '0' ('0'..'7')+ IntegerTypeSuffix? ;

fragment
HexDigit : ('0'..'9'|'a'..'f'|'A'..'F') ;

fragment
IntegerTypeSuffix : ('l'|'L') ;

FloatingPointLiteral
    :   ('0'..'9')+ '.' ('0'..'9')* Exponent? FloatTypeSuffix?
    |   '.' ('0'..'9')+ Exponent? FloatTypeSuffix?
    |   ('0'..'9')+ Exponent FloatTypeSuffix?
    |   ('0'..'9')+ FloatTypeSuffix
    ;

fragment
Exponent : ('e'|'E') ('+'|'-')? ('0'..'9')+ ;

fragment
FloatTypeSuffix : ('f'|'F'|'d'|'D') ;

CharacterLiteral
    :   '\'' ( EscapeSequence | ~('\''|'\\') ) '\''
    ;

StringLiteral
    :  '"' ( EscapeSequence | ~('\\'|'"') )* '"'
    ;

fragment
EscapeSequence
    :   '\\' ('b'|'t'|'n'|'f'|'r'|'\"'|'\''|'\\')
    |   UnicodeEscape
    |   OctalEscape
    ;

fragment
OctalEscape
    :   '\\' ('0'..'3') ('0'..'7') ('0'..'7')
    |   '\\' ('0'..'7') ('0'..'7')
    |   '\\' ('0'..'7')
    ;

fragment
UnicodeEscape
    :   '\\' 'u' HexDigit HexDigit HexDigit HexDigit
    ;
    
Identifier 
    :   Letter (Letter|JavaIDDigit)*
    ;
/**I found this char range in JavaCC's grammar, but Letter and Digit overlap.
   Still works, but...
 */
fragment
Letter
    :  '\u0024' |
       '\u0041'..'\u005a' |
       '\u005f' |
       '\u0061'..'\u007a' |
       '\u00c0'..'\u00d6' |
       '\u00d8'..'\u00f6' |
       '\u00f8'..'\u00ff' |
       '\u0100'..'\u1fff' |
       '\u3040'..'\u318f' |
       '\u3300'..'\u337f' |
       '\u3400'..'\u3d2d' |
       '\u4e00'..'\u9fff' |
       '\uf900'..'\ufaff'
    ;

fragment
JavaIDDigit
    :  '\u0030'..'\u0039' |
       '\u0660'..'\u0669' |
       '\u06f0'..'\u06f9' |
       '\u0966'..'\u096f' |
       '\u09e6'..'\u09ef' |
       '\u0a66'..'\u0a6f' |
       '\u0ae6'..'\u0aef' |
       '\u0b66'..'\u0b6f' |
       '\u0be7'..'\u0bef' |
       '\u0c66'..'\u0c6f' |
       '\u0ce6'..'\u0cef' |
       '\u0d66'..'\u0d6f' |
       '\u0e50'..'\u0e59' |
       '\u0ed0'..'\u0ed9' |
       '\u1040'..'\u1049'
   ;
   
CMT: '/*'(options {greedy=false;}:.)* '*/'{$channel = HIDDEN;};
LCMT: '//' ~('\n'|'\r')* '\r'? '\n' {$channel = HIDDEN;};
WS:  (' '|'\t'|'\r'|'\n')+ {$channel = HIDDEN;};
