package at.ac.tuwien.dbai.alternation.examples;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
/**
 * A directed graph
 *
 * @author Wolfgang Dvorak
 *
 * @param <NTyp> type of the content saved in graph nodes
 */
public class BipartiteGraph<NTyp> implements Iterable<NTyp>{
	/**
	 * Node in a directed graph
	 *
	 * @author Wolfgang Dvorak
	 *
	 * @param <NTyp> type of the content saved in graph nodes
	 */
	public static class Node<NTyp>{
		private NTyp object;
		private Set<Node<NTyp>> parents;
		private Set<Node<NTyp>> childs;

		public Node(NTyp object){
			this.object=object;
			this.parents=new HashSet<Node<NTyp>>();
			this.childs=new HashSet<Node<NTyp>>();
		}

		public NTyp getObject(){
			return this.object;
		}

		public Set<Node<NTyp>> getChildren(){
			return this.childs;
		}

		public Set<Node<NTyp>> getParents(){
			return this.parents;
		}

		public boolean isLinked(){
			return !(this.parents.isEmpty() && this.childs.isEmpty());
		}

		public boolean addParent(Node<NTyp> parent){
			if(!this.parents.contains(parent)){
				this.parents.add(parent);
				return true;
			}else{
				return false;
			}
		}
		public boolean addChild(Node<NTyp> child){
			if(!this.childs.contains(child)){
				this.childs.add(child);
				return true;
			}else{
				return false;
			}
		}

		public boolean removeParent(Node<NTyp> parent){
			return this.parents.remove(parent);
		}

		public boolean removeChild(Node<NTyp> child){
			return this.childs.remove(child);
		}

		public Iterator<Node<NTyp>> parentIterator(){
			return this.parents.iterator();
		}

		public Iterator<Node<NTyp>> childIterator(){
			return this.childs.iterator();
		}

		public String toString() {
			return object.toString();
		}
	}

	private Map<NTyp,Node<NTyp>> nodeSetA;

	private Map<NTyp,Node<NTyp>> nodeSetB;

	public BipartiteGraph() {
		this.nodeSetA=new HashMap<NTyp,Node<NTyp>>();
		this.nodeSetB=new HashMap<NTyp,Node<NTyp>>();
	}

	public Iterator<NTyp> iterator() {
		return null;
	}

	public Iterator<NTyp> iteratorSetA() {
		return nodeSetA.keySet().iterator();
	}
	public Iterator<NTyp> iteratorSetB() {
		return nodeSetB.keySet().iterator();
	}

	public boolean addNodeToSetA(NTyp node){
		if (nodeSetA.containsKey(node)){
			return false;
		}
		nodeSetA.put(node, new Node<NTyp>(node));
		return true;
	}

	public boolean addNodeToSetB(NTyp node){
		if (nodeSetB.containsKey(node)){
			return false;
		}
		nodeSetB.put(node, new Node<NTyp>(node));
		return true;
	}

	public boolean removeNode(NTyp node){
		if (nodeSetA.containsKey(node)){
			Node<NTyp> mnode = nodeSetA.get(node);
			nodeSetA.remove(node);
			// remove edges
			for(Node<NTyp> mnode2 : nodeSetB.values()){
				mnode2.removeChild(mnode);
				mnode2.removeChild(mnode);
			}
			return true;
		}

		else if(nodeSetB.containsKey(node)){
			Node<NTyp> mnode = nodeSetB.get(node);
			nodeSetB.remove(node);
			// remove edges
			for(Node<NTyp> mnode2 : nodeSetA.values()){
				mnode2.removeChild(mnode);
				mnode2.removeChild(mnode);
			}
			return true;
		}
		else return false;
	}

	public boolean addEdge(NTyp node1,NTyp node2){
		if (nodeSetA.containsKey(node1) & nodeSetB.containsKey(node2)){
			nodeSetA.get(node1).addChild(nodeSetB.get(node2));
			nodeSetB.get(node2).addParent(nodeSetA.get(node1));
			return true;
		} else if (nodeSetB.containsKey(node1) & nodeSetA.containsKey(node2)){
			nodeSetB.get(node1).addChild(nodeSetA.get(node2));
			nodeSetA.get(node2).addParent(nodeSetB.get(node1));
			return true;
		}else return false;
	}

	public boolean removeEdge(NTyp node1,NTyp node2){
		if (nodeSetA.containsKey(node1) & nodeSetB.containsKey(node2)){
			return nodeSetA.get(node1).removeChild(nodeSetB.get(node2)) &
			nodeSetB.get(node2).removeParent(nodeSetA.get(node1));
		} else if (nodeSetB.containsKey(node1) & nodeSetA.containsKey(node2)){
			return nodeSetB.get(node1).removeChild(nodeSetA.get(node2)) &
			nodeSetA.get(node2).removeParent(nodeSetB.get(node1));
		}else return false;
	}

	public Node<NTyp> get(NTyp node) {
		if (nodeSetA.containsKey(node)){
			return nodeSetA.get(node);
		} else if (nodeSetB.containsKey(node)){
			return nodeSetB.get(node);
		}else return null;
	}

	/**
	 *
	 * @return number of nodes
	 */
	public int size() {
		return nodeSetA.size()+nodeSetB.size();
	}
}
